from .constants import VALUE, FEATURE_FLAGS
from .overridedatasource import OverrideDataSource


class LocalDictionaryDataSource(OverrideDataSource):
    def __init__(self, source, override_behaviour):
        OverrideDataSource.__init__(self, override_behaviour=override_behaviour)
        dictionary = {}
        for key, value in source.items():
            dictionary[key] = {VALUE: value}
        self._settings = {FEATURE_FLAGS: dictionary}

    def get_overrides(self):
        return self._settings



from .interfaces import ConfigCache


class InMemoryConfigCache(ConfigCache):

    def __init__(self):
        self._value = {}

    def get(self, key):
        return self._value.get(key)

    def set(self, key, value):
        self._value[key] = value



from .constants import VALUE, FEATURE_FLAGS
from .overridedatasource import OverrideDataSource
import json
import os
import logging
import sys

log = logging.getLogger(sys.modules[__name__].__name__)


class LocalFileDataSource(OverrideDataSource):
    def __init__(self, file_path, override_behaviour):
        OverrideDataSource.__init__(self, override_behaviour=override_behaviour)
        if not os.path.exists(file_path):
            log.error('The file \'%s\' does not exists.' % file_path)

        self._file_path = file_path
        self._settings = None
        self._cached_file_stamp = 0

    def get_overrides(self):
        self._reload_file_content()
        return self._settings

    def _reload_file_content(self):
        try:
            stamp = os.stat(self._file_path).st_mtime
            if stamp != self._cached_file_stamp:
                self._cached_file_stamp = stamp
                with open(self._file_path) as file:
                    data = json.load(file)
                    if 'flags' in data:
                        dictionary = {}
                        source = data['flags']
                        for key, value in source.items():
                            dictionary[key] = {VALUE: value}
                        self._settings = {FEATURE_FLAGS: dictionary}
                    else:
                        self._settings = data
        except OSError as e:
            log.error('Could not read the content of the file %s. %s' % (self._file_path, e))
        except ValueError as e:
            log.error('Could not decode json from file %s. %s' % (self._file_path, e))




CONFIG_FILE_NAME = 'config_v5'

PREFERENCES = 'p'
BASE_URL = 'u'
REDIRECT = 'r'

FEATURE_FLAGS = 'f'
VALUE = 'v'
COMPARATOR = 't'
COMPARISON_ATTRIBUTE = 'a'
COMPARISON_VALUE = 'c'
ROLLOUT_PERCENTAGE_ITEMS = 'p'
PERCENTAGE = 'p'
ROLLOUT_RULES = 'r'
VARIATION_ID = 'i'



from enum import IntEnum


class DataGovernance(IntEnum):
    """
    Control the location of the config.json files containing your feature flags
    and settings within the ConfigCat CDN. \n
    Global: Select this if your feature flags are published to all global CDN nodes. \n
    EuOnly: Select this if your feature flags are published to CDN nodes only in the EU.
    """
    Global = 0
    EuOnly = 1



from .constants import FEATURE_FLAGS, ROLLOUT_RULES, VARIATION_ID, VALUE, ROLLOUT_PERCENTAGE_ITEMS, CONFIG_FILE_NAME
from .interfaces import ConfigCatClientException
from .lazyloadingcachepolicy import LazyLoadingCachePolicy
from .manualpollingcachepolicy import ManualPollingCachePolicy
from .autopollingcachepolicy import AutoPollingCachePolicy
from .configfetcher import ConfigFetcher
from .configcache import InMemoryConfigCache
from .datagovernance import DataGovernance
from .overridedatasource import OverrideBehaviour
from .rolloutevaluator import RolloutEvaluator
import logging
import sys
import hashlib
from collections import namedtuple
import copy

log = logging.getLogger(sys.modules[__name__].__name__)

KeyValue = namedtuple('KeyValue', 'key value')


class ConfigCatClient(object):
    sdk_keys = []

    def __init__(self,
                 sdk_key,
                 poll_interval_seconds=60,
                 max_init_wait_time_seconds=5,
                 on_configuration_changed_callback=None,
                 cache_time_to_live_seconds=60,
                 config_cache_class=None,
                 base_url=None,
                 proxies=None,
                 proxy_auth=None,
                 connect_timeout=10,
                 read_timeout=30,
                 flag_overrides=None,
                 data_governance=DataGovernance.Global):

        if sdk_key is None:
            raise ConfigCatClientException('SDK Key is required.')

        if sdk_key in ConfigCatClient.sdk_keys:
            log.warning('A ConfigCat Client is already initialized with sdk_key %s. '
                        'We strongly recommend you to use the ConfigCat Client as '
                        'a Singleton object in your application.' % sdk_key)
        else:
            ConfigCatClient.sdk_keys.append(sdk_key)

        self._sdk_key = sdk_key
        self._override_data_source = flag_overrides
        self._rollout_evaluator = RolloutEvaluator()

        if config_cache_class:
            self._config_cache = config_cache_class()
        else:
            self._config_cache = InMemoryConfigCache()

        if self._override_data_source and self._override_data_source.get_behaviour() == OverrideBehaviour.LocalOnly:
            self._config_fetcher = None
            self._cache_policy = None
        else:
            if poll_interval_seconds > 0:
                self._config_fetcher = ConfigFetcher(sdk_key, 'a', base_url, proxies, proxy_auth, connect_timeout, read_timeout, data_governance)
                self._cache_policy = AutoPollingCachePolicy(self._config_fetcher, self._config_cache,
                                                            self.__get_cache_key(),
                                                            poll_interval_seconds, max_init_wait_time_seconds,
                                                            on_configuration_changed_callback)
            elif cache_time_to_live_seconds > 0:
                self._config_fetcher = ConfigFetcher(sdk_key, 'l', base_url, proxies, proxy_auth, connect_timeout, read_timeout, data_governance)
                self._cache_policy = LazyLoadingCachePolicy(self._config_fetcher, self._config_cache,
                                                            self.__get_cache_key(),
                                                            cache_time_to_live_seconds)
            else:
                self._config_fetcher = ConfigFetcher(sdk_key, 'm', base_url, proxies, proxy_auth, connect_timeout, read_timeout, data_governance)
                self._cache_policy = ManualPollingCachePolicy(self._config_fetcher, self._config_cache,
                                                              self.__get_cache_key())

    def get_value(self, key, default_value, user=None):
        config = self.__get_settings()
        if config is None:
            log.warning('Evaluating get_value(\'%s\') failed. Cache is empty. '
                        'Returning default_value in your get_value call: [%s].' %
                        (key, str(default_value)))
            return default_value

        value, variation_id = self._rollout_evaluator.evaluate(key, user, default_value, None, config)
        return value

    def get_all_keys(self):
        config = self.__get_settings()
        if config is None:
            return []

        feature_flags = config.get(FEATURE_FLAGS, None)
        if feature_flags is None:
            return []

        return list(feature_flags)

    def get_variation_id(self, key, default_variation_id, user=None):
        config = self.__get_settings()
        if config is None:
            log.warning('Evaluating get_variation_id(\'%s\') failed. Cache is empty. '
                        'Returning default_variation_id in your get_variation_id call: [%s].' %
                        (key, str(default_variation_id)))
            return default_variation_id

        value, variation_id = self._rollout_evaluator.evaluate(key, user, None, default_variation_id, config)
        return variation_id

    def get_all_variation_ids(self, user=None):
        keys = self.get_all_keys()
        variation_ids = []
        for key in keys:
            variation_id = self.get_variation_id(key, None, user)
            if variation_id is not None:
                variation_ids.append(variation_id)

        return variation_ids

    def get_key_and_value(self, variation_id):
        config = self.__get_settings()
        if config is None:
            log.warning('Evaluating get_key_and_value(\'%s\') failed. Cache is empty. '
                        'Returning None.' % variation_id)
            return None

        feature_flags = config.get(FEATURE_FLAGS, None)
        if feature_flags is None:
            log.warning('Evaluating get_key_and_value(\'%s\') failed. Cache is empty. '
                        'Returning None.' % variation_id)
            return None

        for key, value in list(feature_flags.items()):
            if variation_id == value.get(VARIATION_ID):
                return KeyValue(key, value[VALUE])

            rollout_rules = value.get(ROLLOUT_RULES, [])
            for rollout_rule in rollout_rules:
                if variation_id == rollout_rule.get(VARIATION_ID):
                    return KeyValue(key, rollout_rule[VALUE])

            rollout_percentage_items = value.get(ROLLOUT_PERCENTAGE_ITEMS, [])
            for rollout_percentage_item in rollout_percentage_items:
                if variation_id == rollout_percentage_item.get(VARIATION_ID):
                    return KeyValue(key, rollout_percentage_item[VALUE])

        log.error('Could not find the setting for the given variation_id: ' + variation_id)
        return None

    def get_all_values(self, user=None):
        keys = self.get_all_keys()
        all_values = {}
        for key in keys:
            value = self.get_value(key, None, user)
            if value is not None:
                all_values[key] = value

        return all_values

    def force_refresh(self):
        if self._cache_policy:
            self._cache_policy.force_refresh()

    def stop(self):
        if self._cache_policy:
            self._cache_policy.stop()
        ConfigCatClient.sdk_keys.remove(self._sdk_key)

    def __get_settings(self):
        if self._override_data_source:
            behaviour = self._override_data_source.get_behaviour()

            if behaviour == OverrideBehaviour.LocalOnly:
                return self._override_data_source.get_overrides()
            elif behaviour == OverrideBehaviour.RemoteOverLocal:
                remote_settings = self._cache_policy.get()
                local_settings = self._override_data_source.get_overrides()
                result = copy.deepcopy(local_settings)
                if FEATURE_FLAGS in remote_settings and FEATURE_FLAGS in local_settings:
                    result[FEATURE_FLAGS].update(remote_settings[FEATURE_FLAGS])
                return result
            elif behaviour == OverrideBehaviour.LocalOverRemote:
                remote_settings = self._cache_policy.get()
                local_settings = self._override_data_source.get_overrides()
                result = copy.deepcopy(remote_settings)
                if FEATURE_FLAGS in remote_settings and FEATURE_FLAGS in local_settings:
                    result[FEATURE_FLAGS].update(local_settings[FEATURE_FLAGS])
                return result

        return self._cache_policy.get()

    def __get_cache_key(self):
        return hashlib.sha1(('python_' + CONFIG_FILE_NAME + '_' + self._sdk_key).encode('utf-8')).hexdigest()



import logging
import sys
from requests import HTTPError

from .readwritelock import ReadWriteLock
from .interfaces import CachePolicy

log = logging.getLogger(sys.modules[__name__].__name__)


class ManualPollingCachePolicy(CachePolicy):
    def __init__(self, config_fetcher, config_cache, cache_key):
        self._config_fetcher = config_fetcher
        self._config_cache = config_cache
        self._cache_key = cache_key
        self._lock = ReadWriteLock()

    def get(self):
        try:
            self._lock.acquire_read()

            config = self._config_cache.get(self._cache_key)
            return config
        finally:
            self._lock.release_read()

    def force_refresh(self):
        force_fetch = False

        try:
            self._lock.acquire_read()
            config = self._config_cache.get(self._cache_key)
            force_fetch = not bool(config)
        finally:
            self._lock.release_read()

        try:
            configuration_response = self._config_fetcher.get_configuration_json(force_fetch)
            if configuration_response.is_fetched():
                configuration = configuration_response.json()
                try:
                    self._lock.acquire_write()
                    self._config_cache.set(self._cache_key, configuration)
                finally:
                    self._lock.release_write()

        except HTTPError as e:
            log.error('Double-check your SDK Key at https://app.configcat.com/sdkkey.'
                      ' Received unexpected response: [%s]' % str(e.response))
        except Exception:
            log.exception(sys.exc_info()[0])

    def stop(self):
        pass



__PREDEFINED__ = ["Identifier", 'Email', 'Country']


class User(object):
    """
     The user object for variation evaluation
    """

    def __init__(self, identifier, email=None, country=None, custom=None):
        self.__identifier = identifier if identifier is not None else ''
        self.__data = {'Identifier': identifier, 'Email': email, 'Country': country}
        self.__custom = custom

    def get_identifier(self):
        return self.__identifier

    def get_attribute(self, attribute):
        attribute = str(attribute)
        if attribute in __PREDEFINED__:
            return self.__data[attribute]

        if self.__custom is not None:
            for customField in self.__custom:
                if customField == attribute:
                    return self.__custom[customField]

        return None

    def __str__(self):
        r = '{\n    "Identifier": "%s"' % self.__identifier
        if self.__data['Email'] is not None:
            r += ',\n    "Email": "%s"' % self.__data['Email']
        if self.__data['Country'] is not None:
            r += ',\n    "Country": "%s"' % self.__data['Country']
        if self.__custom is not None:
            r += ',\n    "Custom": {'
            for customField in self.__custom:
                r += '\n        "%s": "%s",' % (customField, self.__custom[customField])
            r += '\n    }'
        r += '\n}'
        return r



from abc import ABCMeta, abstractmethod


class ConfigCache(object):
    """
        Config cache interface
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def get(self, key):
        """
        :returns the config json object from the cache
        """

    @abstractmethod
    def set(self, key, value):
        """
        Sets the config json cache.
        """


class CachePolicy(object):
    """
        Config cache interface
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def get(self):
        """
        :returns the config json object from the cache
        """

    @abstractmethod
    def force_refresh(self):
        """

        :return:
        """

    @abstractmethod
    def stop(self):
        """

        :return:
        """


class ConfigCatClientException(Exception):
    """
    Generic ConfigCatClientException
    """



import logging
import sys
import datetime
import time
from threading import Thread, Event
from requests import HTTPError
from requests import Timeout

from .readwritelock import ReadWriteLock
from .interfaces import CachePolicy

log = logging.getLogger(sys.modules[__name__].__name__)


class AutoPollingCachePolicy(CachePolicy):

    def __init__(self, config_fetcher, config_cache, cache_key,
                 poll_interval_seconds=60, max_init_wait_time_seconds=5,
                 on_configuration_changed_callback=None):
        if poll_interval_seconds < 1:
            poll_interval_seconds = 1
        if max_init_wait_time_seconds < 0:
            max_init_wait_time_seconds = 0

        self._config_fetcher = config_fetcher
        self._config_cache = config_cache
        self._cache_key = cache_key
        self._poll_interval_seconds = poll_interval_seconds
        self._max_init_wait_time_seconds = datetime.timedelta(seconds=max_init_wait_time_seconds)
        self._on_configuration_changed_callback = on_configuration_changed_callback
        self._initialized = False
        self._is_running = False
        self._start_time = datetime.datetime.utcnow()
        self._lock = ReadWriteLock()

        self.thread = Thread(target=self._run, args=[])
        self.thread.daemon = True
        self._is_started = Event()
        self.thread.start()
        self._is_started.wait()

    def _run(self):
        self._is_running = True
        self._is_started.set()
        while True:
            self.force_refresh()
            time.sleep(self._poll_interval_seconds)
            if not self._is_running:
                break

    def get(self):
        while not self._initialized \
                and datetime.datetime.utcnow() < self._start_time + self._max_init_wait_time_seconds:
            time.sleep(.500)

        try:
            self._lock.acquire_read()
            return self._config_cache.get(self._cache_key)
        finally:
            self._lock.release_read()

    def force_refresh(self):
        try:
            old_configuration = None
            force_fetch = False

            try:
                self._lock.acquire_read()
                old_configuration = self._config_cache.get(self._cache_key)
                force_fetch = not bool(old_configuration)
            finally:
                self._lock.release_read()

            configuration_response = self._config_fetcher.get_configuration_json(force_fetch)

            if configuration_response.is_fetched():
                configuration = configuration_response.json()
                if configuration != old_configuration:
                    try:
                        self._lock.acquire_write()
                        self._config_cache.set(self._cache_key, configuration)
                        self._initialized = True
                    finally:
                        self._lock.release_write()

                    try:
                        if self._on_configuration_changed_callback is not None:
                            self._on_configuration_changed_callback()
                    except Exception:
                        log.exception(sys.exc_info()[0])

            if not self._initialized and old_configuration is not None:
                self._initialized = True
        except HTTPError as e:
            log.error('Double-check your SDK Key at https://app.configcat.com/sdkkey.'
                      ' Received unexpected response: %s' % str(e.response))
        except Timeout:
            log.exception('Request timed out. Timeout values: [connect: {}s, read: {}s]'.format(
                self._config_fetcher.get_connect_timeout(), self._config_fetcher.get_read_timeout()))
        except Exception:
            log.exception(sys.exc_info()[0])

    def stop(self):
        self._is_running = False



import logging
import sys
import datetime
from requests import HTTPError

from .readwritelock import ReadWriteLock
from .interfaces import CachePolicy

log = logging.getLogger(sys.modules[__name__].__name__)


class LazyLoadingCachePolicy(CachePolicy):
    def __init__(self, config_fetcher, config_cache, cache_key, cache_time_to_live_seconds=60):
        if cache_time_to_live_seconds < 1:
            cache_time_to_live_seconds = 1
        self._config_fetcher = config_fetcher
        self._config_cache = config_cache
        self._cache_key = cache_key
        self._cache_time_to_live = datetime.timedelta(seconds=cache_time_to_live_seconds)
        self._lock = ReadWriteLock()
        self._last_updated = None

    def get(self):
        config = None

        try:
            self._lock.acquire_read()

            config = self._config_cache.get(self._cache_key)

            utc_now = datetime.datetime.utcnow()
            if self._last_updated is not None and self._last_updated + self._cache_time_to_live > utc_now:
                if config is not None:
                    return config
        finally:
            self._lock.release_read()

        try:
            self._lock.acquire_write()
            # If while waiting to acquire the write lock another
            # thread has updated the content, then don't bother requesting
            # to the server to minimise time.
            if config is None or self._last_updated is None or self._last_updated + self._cache_time_to_live <= datetime.datetime.utcnow():
                force_fetch = not bool(config)
                self._force_refresh(force_fetch)
        finally:
            self._lock.release_write()

        try:
            self._lock.acquire_read()
            config = self._config_cache.get(self._cache_key)
            return config
        finally:
            self._lock.release_read()

    def force_refresh(self):
        force_fetch = False

        try:
            self._lock.acquire_read()
            config = self._config_cache.get(self._cache_key)
            force_fetch = not bool(config)
        finally:
            self._lock.release_read()

        try:
            self._lock.acquire_write()
            self._force_refresh(force_fetch)
        finally:
            self._lock.release_write()

    def _force_refresh(self, force_fetch):
        try:
            configuration_response = self._config_fetcher.get_configuration_json(force_fetch)
            # set _last_updated regardless of whether the cache is updated
            # or whether a 304 not modified has been sent back as the content
            # we have hasn't been updated on the server so not need
            # for subsequent requests to retry this within the cache time to live
            self._last_updated = datetime.datetime.utcnow()
            if configuration_response.is_fetched():
                configuration = configuration_response.json()
                self._config_cache.set(self._cache_key, configuration)
        except HTTPError as e:
            log.error('Double-check your SDK Key at https://app.configcat.com/sdkkey.'
                      ' Received unexpected response: %s' % str(e.response))
        except Exception:
            log.exception(sys.exc_info()[0])

    def stop(self):
        pass



from setuptools import setup


def parse_requirements(filename):
    lines = (line.strip() for line in open(filename))
    return [line for line in lines if line]


configcatclient_version = '6.0.3'

requirements = parse_requirements('requirements.txt')

setup(
    name='configcat-client',
    version=configcatclient_version,
    packages=['configcatclient'],
    url='https://github.com/configcat/python-sdk',
    license='MIT',
    author='ConfigCat',
    author_email='developer@configcat.com',
    description='ConfigCat SDK for Python. https://configcat.com',
    long_description='Feature Flags created by developers for developers with <3. ConfigCat lets you manage '
                     'feature flags across frontend, backend, mobile, and desktop apps without (re)deploying code. '
                     '% rollouts, user targeting, segmentation. Feature toggle SDKs for all main languages. '
                     'Alternative to LaunchDarkly. '
                     'Host yourself, or use the hosted management app at https://configcat.com.',
    install_requires=requirements,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
    ],
)



from abc import ABCMeta, abstractmethod
from enum import IntEnum


class OverrideBehaviour(IntEnum):
    # When evaluating values, the SDK will not use feature flags & settings from the ConfigCat CDN, but it will use
    # all feature flags & settings that are loaded from local-override sources.
    LocalOnly = 0,

    # When evaluating values, the SDK will use all feature flags & settings that are downloaded from the ConfigCat CDN,
    # plus all feature flags & settings that are loaded from local-override sources. If a feature flag or a setting is
    # defined both in the fetched and the local-override source then the local-override version will take precedence.
    LocalOverRemote = 1,

    # When evaluating values, the SDK will use all feature flags & settings that are downloaded from the ConfigCat CDN,
    # plus all feature flags & settings that are loaded from local-override sources. If a feature flag or a setting is
    # defined both in the fetched and the local-override source then the fetched version will take precedence.
    RemoteOverLocal = 2


class OverrideDataSource(object):
    __metaclass__ = ABCMeta

    def __init__(self, override_behaviour):
        self._override_behaviour = override_behaviour

    def get_behaviour(self):
        return self._override_behaviour

    @abstractmethod
    def get_overrides(self):
        """
        :returns the override dictionary
        """



import hashlib
import semver
import logging
import sys
from .constants import *

from .user import User

log = logging.getLogger(sys.modules[__name__].__name__)


class RolloutEvaluator(object):
    SEMANTIC_VERSION_COMPARATORS = ['<', '<=', '>', '>=']
    COMPARATOR_TEXTS = [
        'IS ONE OF',
        'IS NOT ONE OF',
        'CONTAINS',
        'DOES NOT CONTAIN',
        'IS ONE OF (SemVer)',
        'IS NOT ONE OF (SemVer)',
        '< (SemVer)',
        '<= (SemVer)',
        '> (SemVer)',
        '>= (SemVer)',
        '= (Number)',
        '<> (Number)',
        '< (Number)',
        '<= (Number)',
        '> (Number)',
        '>= (Number)',
        'IS ONE OF (Sensitive)',
        'IS NOT ONE OF (Sensitive)'
    ]

    def evaluate(self, key, user, default_value, default_variation_id, config):
        feature_flags = config.get(FEATURE_FLAGS, None)
        if feature_flags is None:
            log.error('Evaluating get_value(\'%s\') failed. Value not found for key \'%s\' '
                      'Returning default_value: [%s].' %
                      (key, key, str(default_value)))
            return default_value, default_variation_id

        setting_descriptor = feature_flags.get(key, None)

        if setting_descriptor is None:
            log.error('Evaluating get_value(\'%s\') failed. Value not found for key \'%s\' '
                      'Returning default_value: [%s]. Here are the available keys: %s' %
                      (key, key, str(default_value), ', '.join(list(feature_flags))))
            return default_value, default_variation_id

        rollout_rules = setting_descriptor.get(ROLLOUT_RULES, [])
        rollout_percentage_items = setting_descriptor.get(ROLLOUT_PERCENTAGE_ITEMS, [])

        if user is not None and type(user) is not User:
            log.warning('Evaluating get_value(\'%s\'). User Object is not an instance of User type.' % key)
            user = None

        if user is None:
            if len(rollout_rules) > 0 or len(rollout_percentage_items) > 0:
                log.warning('Evaluating get_value(\'%s\'). UserObject missing! '
                            'You should pass a UserObject to get_value(), '
                            'in order to make targeting work properly. '
                            'Read more: https://configcat.com/docs/advanced/user-object/' %
                            key)
            return_value = setting_descriptor.get(VALUE, default_value)
            return_variation_id = setting_descriptor.get(VARIATION_ID, default_variation_id)
            log.info('Returning [%s]' % str(return_value))
            return return_value, return_variation_id

        log_entries = ['Evaluating get_value(\'%s\').' % key, 'User object:\n%s' % str(user)]

        try:
            # Evaluate targeting rules
            for rollout_rule in rollout_rules:
                comparison_attribute = rollout_rule.get(COMPARISON_ATTRIBUTE)
                comparison_value = rollout_rule.get(COMPARISON_VALUE)
                comparator = rollout_rule.get(COMPARATOR)

                user_value = user.get_attribute(comparison_attribute)
                if user_value is None or not user_value:
                    log_entries.append(self._format_no_match_rule(comparison_attribute, user_value, comparator, comparison_value))
                    continue

                value = rollout_rule.get(VALUE)
                variation_id = rollout_rule.get(VARIATION_ID, default_variation_id)

                # IS ONE OF
                if comparator == 0:
                    if str(user_value) in [x.strip() for x in str(comparison_value).split(',')]:
                        log_entries.append(self._format_match_rule(comparison_attribute, user_value, comparator,
                                                                   comparison_value, value))
                        return value, variation_id
                # IS NOT ONE OF
                elif comparator == 1:
                    if str(user_value) not in [x.strip() for x in str(comparison_value).split(',')]:
                        log_entries.append(self._format_match_rule(comparison_attribute, user_value, comparator,
                                                                   comparison_value, value))
                        return value, variation_id
                # CONTAINS
                elif comparator == 2:
                    if str(user_value).__contains__(str(comparison_value)):
                        log_entries.append(self._format_match_rule(comparison_attribute, user_value, comparator,
                                                                   comparison_value, value))
                        return value, variation_id
                # DOES NOT CONTAIN
                elif comparator == 3:
                    if not str(user_value).__contains__(str(comparison_value)):
                        log_entries.append(self._format_match_rule(comparison_attribute, user_value, comparator,
                                                                   comparison_value, value))
                        return value, variation_id

                # IS ONE OF, IS NOT ONE OF (Semantic version)
                elif 4 <= comparator <= 5:
                    try:
                        match = False
                        for x in filter(None, [x.strip() for x in str(comparison_value).split(',')]):
                            match = semver.match(str(user_value).strip(), '==' + x) or match
                        if (match and comparator == 4) or (not match and comparator == 5):
                            log_entries.append(self._format_match_rule(comparison_attribute, user_value, comparator,
                                                                       comparison_value, value))
                            return value, variation_id
                    except ValueError as e:
                        message = self._format_validation_error_rule(comparison_attribute, user_value, comparator,
                                                                     comparison_value, str(e))
                        log.warning(message)
                        log_entries.append(message)
                        continue

                # LESS THAN, LESS THAN OR EQUALS TO, GREATER THAN, GREATER THAN OR EQUALS TO (Semantic version)
                elif 6 <= comparator <= 9:
                    try:
                        if semver.match(str(user_value).strip(),
                                        self.SEMANTIC_VERSION_COMPARATORS[comparator - 6] + str(comparison_value).strip()):
                            log_entries.append(self._format_match_rule(comparison_attribute, user_value, comparator,
                                                                       comparison_value, value))
                            return value, variation_id
                    except ValueError as e:
                        message = self._format_validation_error_rule(comparison_attribute, user_value, comparator,
                                                                     comparison_value, str(e))
                        log.warning(message)
                        log_entries.append(message)
                        continue
                elif 10 <= comparator <= 15:
                    try:
                        user_value_float = float(str(user_value).replace(",", "."))
                        comparison_value_float = float(str(comparison_value).replace(",", "."))

                        if (comparator == 10 and user_value_float == comparison_value_float) \
                                or (comparator == 11 and user_value_float != comparison_value_float) \
                                or (comparator == 12 and user_value_float < comparison_value_float) \
                                or (comparator == 13 and user_value_float <= comparison_value_float) \
                                or (comparator == 14 and user_value_float > comparison_value_float) \
                                or (comparator == 15 and user_value_float >= comparison_value_float):
                            log_entries.append(self._format_match_rule(comparison_attribute, user_value, comparator,
                                                                       comparison_value, value))
                            return value, variation_id
                    except Exception as e:
                        message = self._format_validation_error_rule(comparison_attribute, user_value, comparator,
                                                                     comparison_value, str(e))
                        log.warning(message)
                        log_entries.append(message)
                        continue
                # IS ONE OF (Sensitive)
                elif comparator == 16:
                    if str(hashlib.sha1(user_value.encode('utf8')).hexdigest()) in [x.strip() for x in str(comparison_value).split(',')]:
                        log_entries.append(self._format_match_rule(comparison_attribute, user_value, comparator,
                                                                   comparison_value, value))
                        return value, variation_id
                # IS NOT ONE OF (Sensitive)
                elif comparator == 17:
                    if str(hashlib.sha1(user_value.encode('utf8')).hexdigest()) not in [x.strip() for x in str(comparison_value).split(',')]:
                        log_entries.append(self._format_match_rule(comparison_attribute, user_value, comparator,
                                                                   comparison_value, value))
                        return value, variation_id

                log_entries.append(self._format_no_match_rule(comparison_attribute, user_value, comparator, comparison_value))

            # Evaluate variations
            if len(rollout_percentage_items) > 0:
                user_key = user.get_identifier()
                hash_candidate = ('%s%s' % (key, user_key)).encode('utf-8')
                hash_val = int(hashlib.sha1(hash_candidate).hexdigest()[:7], 16) % 100

                bucket = 0
                for rollout_percentage_item in rollout_percentage_items or []:
                    bucket += rollout_percentage_item.get(PERCENTAGE, 0)
                    if hash_val < bucket:
                        percentage_value = rollout_percentage_item.get(VALUE)
                        variation_id = rollout_percentage_item.get(VARIATION_ID, default_variation_id)
                        log_entries.append('Evaluating %% options. Returning %s' % percentage_value)
                        return percentage_value, variation_id

            return_value = setting_descriptor.get(VALUE, default_value)
            return_variation_id = setting_descriptor.get(VARIATION_ID, default_variation_id)
            log_entries.append('Returning %s' % return_value)
            return return_value, return_variation_id
        finally:
            log.info('\n'.join(log_entries))

    def _format_match_rule(self, comparison_attribute, user_value, comparator, comparison_value, value):
        return 'Evaluating rule: [%s:%s] [%s] [%s] => match, returning: %s' \
               % (comparison_attribute, user_value, self.COMPARATOR_TEXTS[comparator], comparison_value, value)

    def _format_no_match_rule(self, comparison_attribute, user_value, comparator, comparison_value):
        return 'Evaluating rule: [%s:%s] [%s] [%s] => no match' \
               % (comparison_attribute, user_value, self.COMPARATOR_TEXTS[comparator], comparison_value)

    def _format_validation_error_rule(self, comparison_attribute, user_value, comparator, comparison_value, error):
        return 'Evaluating rule: [%s:%s] [%s] [%s] => SKIP rule. Validation error: %s' \
               % (comparison_attribute, user_value, self.COMPARATOR_TEXTS[comparator], comparison_value, error)



import requests
import logging
import sys
from enum import IntEnum

from .datagovernance import DataGovernance
from .version import CONFIGCATCLIENT_VERSION
from .constants import *

if sys.version_info < (2, 7, 9):
    requests.packages.urllib3.disable_warnings()

BASE_URL_GLOBAL = 'https://cdn-global.configcat.com'
BASE_URL_EU_ONLY = 'https://cdn-eu.configcat.com'
BASE_PATH = 'configuration-files/'
BASE_EXTENSION = '/' + CONFIG_FILE_NAME + '.json'

log = logging.getLogger(sys.modules[__name__].__name__)


class RedirectMode(IntEnum):
    NoRedirect = 0,
    ShouldRedirect = 1,
    ForceRedirect = 2


class FetchResponse(object):
    def __init__(self, response):
        self._response = response

    def json(self):
        """Returns the json-encoded content of a response, if any.
        :raises ValueError: If the response body does not contain valid json.
        """
        return self._response.json()

    def is_fetched(self):
        """Gets whether a new configuration value was fetched or not.
        :return: True if a new configuration value was fetched, otherwise false.
        """
        return 200 <= self._response.status_code < 300

    def is_not_modified(self):
        """Gets whether the fetch resulted a '304 Not Modified' or not.
        :return: True if the fetch resulted a '304 Not Modified' code, otherwise false.
        """
        return self._response.status_code == 304


class ConfigFetcher(object):
    def __init__(self, sdk_key, mode, base_url=None, proxies=None, proxy_auth=None,
                 connect_timeout=10, read_timeout=30, data_governance=DataGovernance.Global):
        self._sdk_key = sdk_key
        self._proxies = proxies
        self._proxy_auth = proxy_auth
        self._connect_timeout = connect_timeout
        self._read_timeout = read_timeout
        self._etag = ''
        self._headers = {'User-Agent': 'ConfigCat-Python/' + mode + '-' + CONFIGCATCLIENT_VERSION,
                         'X-ConfigCat-UserAgent': 'ConfigCat-Python/' + mode + '-' + CONFIGCATCLIENT_VERSION,
                         'Content-Type': "application/json"}
        if base_url is not None:
            self._base_url_overridden = True
            self._base_url = base_url.rstrip('/')
        else:
            self._base_url_overridden = False
            if data_governance == DataGovernance.EuOnly:
                self._base_url = BASE_URL_EU_ONLY
            else:
                self._base_url = BASE_URL_GLOBAL

    def get_connect_timeout(self):
        return self._connect_timeout

    def get_read_timeout(self):
        return self._read_timeout

    def get_configuration_json(self, force_fetch=False, retries=0):
        """
        :return: Returns the FetchResponse object contains configuration json Dictionary
        """
        uri = self._base_url + '/' + BASE_PATH + self._sdk_key + BASE_EXTENSION
        headers = self._headers
        if self._etag and not force_fetch:
            headers['If-None-Match'] = self._etag
        else:
            headers['If-None-Match'] = None

        response = requests.get(uri, headers=headers, timeout=(self._connect_timeout, self._read_timeout),
                                proxies=self._proxies, auth=self._proxy_auth)
        response.raise_for_status()
        etag = response.headers.get('Etag')
        if etag:
            self._etag = etag

        fetch_response = FetchResponse(response)

        # If there wasn't a config change, we return the response.
        if not fetch_response.is_fetched():
            return fetch_response

        preferences = fetch_response.json().get(PREFERENCES, None)
        if preferences is None:
            return fetch_response

        base_url = preferences.get(BASE_URL)

        # If the base_url is the same as the last called one, just return the response.
        if base_url is None or self._base_url == base_url:
            return fetch_response

        redirect = preferences.get(REDIRECT)
        # If the base_url is overridden, and the redirect parameter is not 2 (force),
        # the SDK should not redirect the calls and it just have to return the response.
        if self._base_url_overridden and redirect != int(RedirectMode.ForceRedirect):
            return fetch_response

        # The next call should use the base_url provided in the config json
        self._base_url = base_url

        # If the redirect property == 0 (redirect not needed), return the response
        if redirect == int(RedirectMode.NoRedirect):
            # Return the response
            return fetch_response

        # Try to download again with the new url

        if redirect == int(RedirectMode.ShouldRedirect):
            log.warning('Your data_governance parameter at ConfigCatClient initialization is not in sync '
                        'with your preferences on the ConfigCat Dashboard: '
                        'https://app.configcat.com/organization/data-governance. '
                        'Only Organization Admins can access this preference.')

        # To prevent loops we check if we retried at least 3 times with the new base_url
        if retries >= 2:
            log.error('Redirect loop during config.json fetch. Please contact support@configcat.com.')
            return fetch_response

        # Retry the config download with the new base_url
        return self.get_configuration_json(force_fetch, retries + 1)



from .configcatclient import ConfigCatClient
from .interfaces import ConfigCatClientException
from .datagovernance import DataGovernance


def create_client(sdk_key, data_governance=DataGovernance.Global):
    """
    Create an instance of ConfigCatClient and setup Auto Poll mode with default options

    :param sdk_key: ConfigCat SDK Key to access your configuration.
    :param data_governance:
    Default: Global. Set this parameter to be in sync with the Data Governance preference on the Dashboard: \n
    https://app.configcat.com/organization/data-governance \n
    (Only Organization Admins have access)
    """
    return create_client_with_auto_poll(sdk_key, data_governance=data_governance)


def create_client_with_auto_poll(sdk_key, poll_interval_seconds=60, max_init_wait_time_seconds=5,
                                 on_configuration_changed_callback=None, config_cache_class=None,
                                 base_url=None, proxies=None, proxy_auth=None, connect_timeout=10, read_timeout=30,
                                 flag_overrides=None,
                                 data_governance=DataGovernance.Global):
    """
    Create an instance of ConfigCatClient and setup Auto Poll mode with custom options

    :param sdk_key: ConfigCat SDK Key to access your configuration.
    :param poll_interval_seconds: The client's poll interval in seconds. Default: 60 seconds.
    :param on_configuration_changed_callback: You can subscribe to configuration changes with this callback
    :param max_init_wait_time_seconds: maximum waiting time for first configuration fetch.
    :param config_cache_class: If you want to use custom caching instead of the client's default InMemoryConfigCache,
    You can provide an implementation of ConfigCache.
    :param base_url: You can set a base_url if you want to use a proxy server between your application and ConfigCat
    :param proxies: Proxy addresses. e.g. { 'https': 'your_proxy_ip:your_proxy_port' }
    :param proxy_auth: Proxy authentication. e.g. HTTPProxyAuth('username', 'password')
    :param connect_timeout: The number of seconds to wait for the server to make the initial connection
    (i.e. completing the TCP connection handshake). Default: 10 seconds.
    :param read_timeout: The number of seconds to wait for the server to respond before giving up. Default: 30 seconds.
    :param flag_overrides: An OverrideDataSource implementation used to override feature flags & settings.
    :param data_governance:
    Default: Global. Set this parameter to be in sync with the Data Governance preference on the Dashboard: \n
    https://app.configcat.com/organization/data-governance \n
    (Only Organization Admins have access)
    """

    if sdk_key is None:
        raise ConfigCatClientException('SDK Key is required.')

    if poll_interval_seconds < 1:
        poll_interval_seconds = 1

    if max_init_wait_time_seconds < 0:
        max_init_wait_time_seconds = 0

    return ConfigCatClient(sdk_key=sdk_key,
                           poll_interval_seconds=poll_interval_seconds,
                           max_init_wait_time_seconds=max_init_wait_time_seconds,
                           on_configuration_changed_callback=on_configuration_changed_callback,
                           cache_time_to_live_seconds=0,
                           config_cache_class=config_cache_class,
                           base_url=base_url,
                           proxies=proxies,
                           proxy_auth=proxy_auth,
                           connect_timeout=connect_timeout,
                           read_timeout=read_timeout,
                           flag_overrides=flag_overrides,
                           data_governance=data_governance)


def create_client_with_lazy_load(sdk_key, cache_time_to_live_seconds=60, config_cache_class=None,
                                 base_url=None, proxies=None, proxy_auth=None, connect_timeout=10, read_timeout=30,
                                 flag_overrides=None,
                                 data_governance=DataGovernance.Global):
    """
    Create an instance of ConfigCatClient and setup Lazy Load mode with custom options

    :param sdk_key: ConfigCat SDK Key to access your configuration.
    :param cache_time_to_live_seconds: The cache TTL.
    :param config_cache_class: If you want to use custom caching instead of the client's default InMemoryConfigCache,
    You can provide an implementation of ConfigCache.
    :param base_url: You can set a base_url if you want to use a proxy server between your application and ConfigCat
    :param proxies: Proxy addresses. e.g. { "https": "your_proxy_ip:your_proxy_port" }
    :param proxy_auth: Proxy authentication. e.g. HTTPProxyAuth('username', 'password')
    :param connect_timeout: The number of seconds to wait for the server to make the initial connection
    (i.e. completing the TCP connection handshake). Default: 10 seconds.
    :param read_timeout: The number of seconds to wait for the server to respond before giving up. Default: 30 seconds.
    :param flag_overrides: An OverrideDataSource implementation used to override feature flags & settings.
    :param data_governance:
    Default: Global. Set this parameter to be in sync with the Data Governance preference on the Dashboard: \n
    https://app.configcat.com/organization/data-governance \n
    (Only Organization Admins have access)
    """

    if sdk_key is None:
        raise ConfigCatClientException('SDK Key is required.')

    if cache_time_to_live_seconds < 1:
        cache_time_to_live_seconds = 1

    return ConfigCatClient(sdk_key=sdk_key,
                           poll_interval_seconds=0,
                           max_init_wait_time_seconds=0,
                           on_configuration_changed_callback=None,
                           cache_time_to_live_seconds=cache_time_to_live_seconds,
                           config_cache_class=config_cache_class,
                           base_url=base_url,
                           proxies=proxies,
                           proxy_auth=proxy_auth,
                           connect_timeout=connect_timeout,
                           read_timeout=read_timeout,
                           flag_overrides=flag_overrides,
                           data_governance=data_governance)


def create_client_with_manual_poll(sdk_key, config_cache_class=None,
                                   base_url=None, proxies=None, proxy_auth=None, connect_timeout=10, read_timeout=30,
                                   flag_overrides=None,
                                   data_governance=DataGovernance.Global):
    """
    Create an instance of ConfigCatClient and setup Manual Poll mode with custom options

    :param sdk_key: ConfigCat SDK Key to access your configuration.
    :param config_cache_class: If you want to use custom caching instead of the client's default InMemoryConfigCache,
    You can provide an implementation of ConfigCache.
    :param base_url: You can set a base_url if you want to use a proxy server between your application and ConfigCat
    :param proxies: Proxy addresses. e.g. { "https": "your_proxy_ip:your_proxy_port" }
    :param proxy_auth: Proxy authentication. e.g. HTTPProxyAuth('username', 'password')
    :param connect_timeout: The number of seconds to wait for the server to make the initial connection
    (i.e. completing the TCP connection handshake). Default: 10 seconds.
    :param read_timeout: The number of seconds to wait for the server to respond before giving up. Default: 30 seconds.
    :param flag_overrides: An OverrideDataSource implementation used to override feature flags & settings.
    :param data_governance:
    Default: Global. Set this parameter to be in sync with the Data Governance preference on the Dashboard: \n
    https://app.configcat.com/organization/data-governance \n
    (Only Organization Admins have access)
    """

    if sdk_key is None:
        raise ConfigCatClientException('SDK Key is required.')

    return ConfigCatClient(sdk_key=sdk_key,
                           poll_interval_seconds=0,
                           max_init_wait_time_seconds=0,
                           on_configuration_changed_callback=None,
                           cache_time_to_live_seconds=0,
                           config_cache_class=config_cache_class,
                           base_url=base_url,
                           proxies=proxies,
                           proxy_auth=proxy_auth,
                           connect_timeout=connect_timeout,
                           read_timeout=read_timeout,
                           flag_overrides=flag_overrides,
                           data_governance=data_governance)
